"""Re-usable layout snippets."""
from ezcharts.layout.snippets.banner import Banner   # noqa: F401,F403
from ezcharts.layout.snippets.grid import Grid   # noqa: F401,F403
from ezcharts.layout.snippets.offcanvas import OffCanvas   # noqa: F401,F403
from ezcharts.layout.snippets.progress import Progress   # noqa: F401,F403
from ezcharts.layout.snippets.stats import Stats  # noqa: F401,F403
from ezcharts.layout.snippets.table import DataTable   # noqa: F401,F403
from ezcharts.layout.snippets.tabs import Tabs   # noqa: F401,F403



"""Run ezCharts demo."""
import argparse

from dominate.tags import p
import numpy as np
import pandas as pd
from pkg_resources import resource_filename

import ezcharts as ezc
from ezcharts import util
from ezcharts.components.ezchart import EZChart
from ezcharts.components.fastcat import SeqSummary
from ezcharts.components.nextclade import NextClade, NXTComponent
from ezcharts.components.reports.labs import LabsReport
from ezcharts.components.theme import LAB_head_resources
from ezcharts.layout.snippets import DataTable
from ezcharts.layout.snippets import Grid
from ezcharts.layout.snippets import OffCanvas
from ezcharts.layout.snippets import Progress
from ezcharts.layout.snippets import Stats
from ezcharts.layout.snippets import Tabs
from ezcharts.layout.snippets.cards import Cards
from ezcharts.layout.snippets.cards import ICard
from ezcharts.plots import Plot
from ezcharts.plots.ideogram import ideogram

# Setup simple globals
WORKFLOW_NAME = 'wf-template'
REPORT_TITLE = f'{WORKFLOW_NAME}-report'


def main(args):
    """Demo for ezCharts functionality."""
    logger = util.get_named_logger("ezChrtDemo")

    # Good location to prepare plots
    logger.info('Building plots')

    # Example data
    raw_df = pd.DataFrame({
        # as laid out in echarts docs for a dataset
        'product': [
            'Matcha Latte', 'Milk Tea', 'Cheese Cocoa', 'Walnut Brownie'],
        '2015': [43.3, 83.1, 86.4, 72.4],
        '2016': [85.8, 73.4, 65.2, 53.9],
        '2017': [93.7, 55.1, 82.5, 39.1]
    })
    # how we'd likely normally have it
    example_df = raw_df.melt(
        id_vars=['product'], value_vars=['2015', '2016', '2017'],
        var_name='year', value_name='sales')

    def example_plot(style="line") -> Plot:
        """Create example plot."""
        if style == "line":
            plot = ezc.lineplot(data=example_df, x='year', y='sales', hue='product')
        elif style == "scatter":
            plot = ezc.scatterplot(data=example_df, x='year', y='sales', hue='product')
        elif style == 'histogram':
            hist_data = pd.DataFrame(
                        [np.random.normal(loc=10.0, size=1000),
                         np.random.normal(loc=12.0, size=1000),
                         np.random.normal(loc=14.0, size=1000)]).T
            hist_data.columns = ['sample1', 'sample2', 'sample3']
            plot = ezc.histplot(hist_data, bins=50, stat='proportion')

            plot.xAxis.name = 'Read length'
            plot.yAxis.name = 'Number of reads'
        elif style == "bar":
            plot = ezc.barplot(data=example_df, x='year', y='sales', hue='product')
        elif style == "heatmap":
            # make data a matrix like you would expect for a heatmap
            df = raw_df.set_index([raw_df.columns[0]])
            # set the names of the axes
            df = df.rename_axis("Product")
            df = df.rename_axis("Year", axis=1)
            plot = ezc.heatmap(data=df)
        else:
            raise ValueError("Unknown plot style")
        plot.title = {"text": f"Example {style} chart"}
        return plot

    # Then, time to construct a report
    logger.info('Building report')

    # Example data
    params = resource_filename('ezcharts', "data/test/params.json")
    versions = resource_filename('ezcharts', "data/test/versions.txt")
    nxt_json = resource_filename('ezcharts', "data/test/nextclade.json")
    fastcat_results = resource_filename(
        'ezcharts', "data/test/fastcat.stats.gz")

    # Create report
    # Note we need to add nextclade as a resource
    report = LabsReport(
        REPORT_TITLE, WORKFLOW_NAME, params, versions,
        head_resources=[*LAB_head_resources, NXTComponent])

    # Add a header badge via a method on the report element.
    report.add_badge('Test badge', bg_class="bg-primary")

    # Make a progress bar for the median accuracy box
    median_accuracy = Progress(
            value_min=0,
            value_max=100,
            value_now=98.9,
            bar_cls="progress-bar-striped",
            height=50)

    # Add something directly to main_content.
    with report.main_content:
        # Stats is a snippet, but we could also just add any html
        # tags here. Snippets are little reusable portions of html,
        # styles and or scripts in python class format. They often
        # expose methods for adding content to them in a simple way.
        # E.g.
        Stats(
            columns=3,
            items=[
                ('1213986', 'Read count'),
                (median_accuracy, 'Median accuracy'),
                ('10213 bp', 'Some other stat')
            ])

    # Add sequence summary component
    with report.add_section('Sequence summaries', 'Summaries'):
        SeqSummary(fastcat_results)

    # This also adds to main_content, but provides a nice
    # container snippet as a starting context.
    with report.add_section('Controls', 'Controls'):
        # This is an example of how to use bootstrap cards
        Cards(
            columns=2,
            items=[
                ICard(
                    head="Positive",
                    body="The positive passed",
                    footer="All postive controls within threshold",
                    classes="bg-success text-white"),
                ICard(
                    head="No Template Control",
                    body="The NTC failed",
                    footer="NTC outside set thresholds.",
                    classes="bg-danger text-white")])
        OffCanvas(
            label="More details (Offcanvas demo)",
            title="Details of Controls",
            body="You can put more details here....")

    # This also adds to main_content, but provides a nice
    # container snippet as a starting context.
    with report.add_section('Alignment results', 'Results'):
        # This is the tabbed section with ezcharts!
        tabs = Tabs()
        with tabs.add_tab('Summary'):
            # Grids are snippets that provide responsive
            # layouts via css grid
            with Grid():
                EZChart(example_plot("line"), 'epi2melabs')
                EZChart(example_plot("scatter"), 'epi2melabs')
                EZChart(example_plot("bar"), 'epi2melabs')
                EZChart(example_plot("histogram"), 'epi2melabs')
        with tabs.add_tab('Accuracy'):
            p("This is a mixed tab!")
            EZChart(example_plot(), 'epi2melabs')
        with tabs.add_tab('Depth'):
            p('Testing, testing, 1 2 3')

        # Dropdowns are nested one more level
        with tabs.add_dropdown_menu('Example', change_header=False):
            with tabs.add_dropdown_tab('First'):
                EZChart(example_plot(), 'epi2melabs')
            with tabs.add_dropdown_tab('Second'):
                p("This is the second dropdown item.")
                # set dodge=False for a stacked barplot
                plot = ezc.barplot(
                    data=example_df, x='year', y='sales', hue='product', dodge=False
                )
                EZChart(plot, 'epi2melabs')

    with report.add_section('Nextclade results', 'Nextclade', True):
        NextClade(nxt_json)

    with report.add_section('Heatmap', 'Heatmap'):
        EZChart(example_plot("heatmap"), 'epi2melabs')

    with report.add_section('Human Genome', 'Genome'):
        EZChart(
            ideogram(blocks='cytobands'),
            'epi2melabs', height="800px")

    with report.add_section('Table', 'Table'):
        tabs = Tabs()
        with tabs.add_tab('Table'):
            df = pd.DataFrame(np.random.random((4, 4))) * 10
            df.columns = pd.MultiIndex.from_product(
                [['sample1', 'sample2'], ['mapped', 'unmapped']])
            df['chr'] = [1, 2, 3, 4]
            df = df.set_index('chr', drop=True)
            DataTable.from_pandas(df, use_index=True)
        with tabs.add_tab('Export'):
            df_simple = pd.DataFrame(np.random.random((4, 2))) * 10
            df_simple.columns = ['sample1', 'sample2']
            DataTable.from_pandas(df_simple, export=True, file_name='samples')

    logger.info('Reticulating splines')
    report.write(args.output)


def argparser():
    """Argument parser for entrypoint."""
    parser = argparse.ArgumentParser(
        "ezcharts demo",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        add_help=False)
    parser.add_argument(
        "--output", default="ezcharts_demo_report.html",
        help="Output HTML file.")
    return parser



"""Simple eCharts API."""

__version__ = "0.3.6"

import argparse
import importlib
import logging

from ezcharts import util
from ezcharts.plots.axisgrid import *  # noqa: F401,F403
from ezcharts.plots.categorical import *  # noqa: F401,F403
from ezcharts.plots.distribution import *  # noqa: F401,F403
from ezcharts.plots.matrix import *  # noqa: F401,F403
from ezcharts.plots.regression import *  # noqa: F401,F403
from ezcharts.plots.relational import *  # noqa: F401,F403


def cli():
    """Run ezcharts entry point."""
    parser = argparse.ArgumentParser(
        'ezcharts',
        parents=[util._log_level()],
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    parser.add_argument(
        '-v', '--version', action='version',
        version='%(prog)s {}'.format(__version__))

    subparsers = parser.add_subparsers(
        title='subcommands', description='valid commands',
        help='additional help', dest='command')
    subparsers.required = True

    # all component demos, plus some others
    components = ['params', 'nextclade', 'fastcat']
    others = ['ezcharts.demo', 'ezcharts.plots', 'ezcharts.plots.ideogram']

    demos = [f'ezcharts.components.{comp}' for comp in components] + others
    for module in demos:
        mod = importlib.import_module(module)
        p = subparsers.add_parser(
            module.split(".")[-1], parents=[mod.argparser()])
        p.set_defaults(func=mod.main)

    args = parser.parse_args()
    logging.basicConfig(
        format='[%(asctime)s - %(name)s] %(message)s',
        datefmt='%H:%M:%S', level=logging.INFO)
    logging.captureWarnings(True)
    logger = logging.getLogger(__package__)
    logger.setLevel(args.log_level)

    args.func(args)



"""Plotting functionality via echarts."""

import argparse

import numpy as np
import pandas as pd
from pkg_resources import resource_filename
import sigfig

from ezcharts import util as ezutil
from ezcharts.components.ezchart import EZChart
from ezcharts.components.reports.comp import ComponentReport
from ezcharts.plots import util
from ezcharts.plots._model import EChartsOption
from ezcharts.plots.util import JSCode

# NOTE: the add_x methods below allow for type checking that pydantic V1 would
#       otherwise not perform, e.g. plt.series.append({...}) evades checking
#       https://github.com/pydantic/pydantic/issues/496

_logger = ezutil.get_named_logger("Plotter")


class AxisLabelFormatter(JSCode):
    """Formatter for echarts axis labels."""

    def __init__(self, sci_notation=(0.0001, 10000)):
        """Init.

        :param: sci_notation:
            True: use scientific notation for the axis.
            False: use raw values for the axis.
            tuple: Upper and lower limits for determining use of scientific notation.
        """
        self.sci_notation = sci_notation

    def apply(self, values=None):
        """
        Create the JS formatter.

        Use the values from the axis to dertermine if scientific notation
        or raw values will be used. If values is None, use raw values as this is
        likely a category axis.

        :param values: The raw values from the current axis.
        """
        self.jscode = """function(value, index){
                        if (value == 0){
                            return 0;
                        }
                        if (@use_sci == 'false'){
                            return value;
                        }
                        if (@use_sci == 'true'){
                            return value.toExponential();
                        }
                     }
        """
        if values is None:
            use_sci = False
        else:
            if self.sci_notation is True:
                use_sci = True
            elif self.sci_notation is False:
                use_sci = False
            else:
                # If any values fall outside the limits, covert all to sci_notation.
                # Do not include zeros as they are always plotted as '0'.
                sci_limits = self.sci_notation
                v = np.array(values)
                v = v[v != 0]
                if ((v < sci_limits[0]) | (v > sci_limits[1])).any():
                    use_sci = True
                else:
                    use_sci = False
        super().__init__(self.jscode.replace('@use_sci', f"'{str(use_sci).lower()}'"))
        return use_sci


class Plot(EChartsOption):
    """Main plotting interface."""

    def __init__(self, *args, **kwargs):
        """Initialize a plot with defaults."""
        super().__init__(*args, **kwargs)
        self.toolbox = {
            "show": True,
            "feature": {
                "dataZoom": {"show": True},
                "dataView": {"readOnly": False},
                "restore": {},
                "saveAsImage": {}
            }
        }
        self.color = util.choose_palette()

    @property
    def logger(self):
        """Return logger for class."""
        return _logger

    def add_series(self, spec):
        """Add a series to chart."""
        orig = self.series
        self.series = [spec]  # validates new entry
        if orig is not None:
            self.series = orig + self.series

    def add_dataset(self, spec):
        """Add a dataset to a chart."""
        orig = self.dataset
        self.dataset = [spec]  # validates new entry
        if orig is not None:
            self.dataset = orig + self.dataset

    def to_json(self, **kwargs):
        """Create a json representation of options.

        Here we clean up the serialised json in the event that it includes
        any javascript code.
        """
        self.finalise()
        return JSCode._clean(self.json(exclude_unset=True))

    def render_html(self, output, **kwargs):
        """Render plot to a file.

        :params output: output file.
        :param kwargs: passed to `EZChart`.
        """
        try:
            title = self.title.text
        except AttributeError:
            title = 'ezChart Plot'
        chart = EZChart(self, 'epi2melabs', **kwargs)
        report = ComponentReport(title, chart)
        report.write(output)

    def finalise(self):
        """Apply a standard set of defaults to patch eCharts.

        Prevent overlap of axis name with labels and other things.
        """
        self.fix_axis_labels()

    def _axes_dimensions(self):
        """Get axes for each dimension."""
        axes = list()
        if not isinstance(self.xAxis, list):
            axes.append([self.xAxis, 0])
        if not isinstance(self.yAxis, list):
            axes.append([self.yAxis, 1])
        return axes

    def fix_axis_labels(self):
        """Try to place axis labels so that they don't overlap tick labels."""
        if hasattr(self, 'grid') and self.grid is not None:
            self.logger.warning(
                "Cannot correct axis labels in complicated scenarios.")
            return

        # we only try to fix up axes if we have one, else we don't
        # really know what's going on
        axes = self._axes_dimensions()

        # to make space for axis labels we shrink the plot by setting a grid
        # and changing the margins of its sole component
        self.grid = dict()

        for axis, data_idx in axes:
            axis.nameLocation = 'middle'  # 'cos eCharts its weird

            # Get all the raw values from each dataset that contains a source.
            # Warning: Just taking the raw datasource here.
            # Any plots with transformed data may not have axes setup correctly.
            raw_vals = []
            for ds in self.dataset:
                if ds.source is not None:
                    # `ds.source` could be a numpy array or just a list of lists; we'll
                    # transform it into a `pd.DataFrame` so that we can do column-wise
                    # indexing while preserving dtypes
                    raw_vals.extend(pd.DataFrame(ds.source).iloc[:, data_idx])

            # Allow formatter to be set by user.
            if axis.axisLabel is None:
                axis.axisLabel = dict(formatter=AxisLabelFormatter())
            elif axis.axisLabel.formatter is None:
                axis.axisLabel.formatter = AxisLabelFormatter()

            if axis.type == 'value':
                is_sci = axis.axisLabel.formatter.apply(raw_vals)
                # If using sci. notation, there will be 4-5 characters in the label
                if is_sci:
                    max_n_label_digits = 5
                else:
                    round_vals = [
                        sigfig.round(val, sigfigs=1) for val in raw_vals]
                    max_n_label_digits = max([len(str(x)) for x in round_vals])
            else:
                axis.axisLabel.formatter.apply()
                max_n_label_digits = max([len(str(x)) for x in raw_vals])

            if axis == self.xAxis:
                name_offset = 25
                try:
                    rotation = axis.axisLabel.rotate
                except AttributeError:
                    rotation = 0
                if rotation != 0:  # Let's assume a sensible rotation of 45
                    # Rotation makes axis labels project downwards more
                    name_offset = 25 + max_n_label_digits * 4
                    self.grid.bottom = name_offset + 15

            elif axis == self.yAxis:
                name_offset = 20 + max_n_label_digits * 6
                self.grid.left = name_offset + 10

            axis.nameGap = name_offset

        return self


class _HistogramPlot(Plot):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _axes_dimensions(self):
        """Get correct axes for histograms.

        For histograms, the first two dataset columns are start/stop
        rectangle x coords. The third column contains bar heights.
        """
        return [self.xAxis, 1], [self.yAxis, 2]


def main(args):
    """Entry point to demonstrate an ezChart."""
    plt = Plot.parse_file(args.plot_spec)
    plt.render_html(args.output)


def argparser():
    """Argument parser for entrypoint."""
    parser = argparse.ArgumentParser(
        'ezChart Demo',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        add_help=False)
    parser.add_argument(
        "--plot_spec",
        default=resource_filename('ezcharts', "data/test/plot-spec.json"),
        help=(
            "A JSON file defining an eCharts plot",
            "key/values."
        ))
    parser.add_argument(
        "--output",
        default="ezcharts_plot_report.html",
        help="Output HTML file.")
    return parser



"""Get default cards layouts."""
from typing import List, NamedTuple, Optional, Type

from dominate.tags import div, h5, html_tag, p

from ezcharts.layout.base import IClasses, Snippet
from ezcharts.layout.snippets import Grid
from ezcharts.layout.util import cls


class ICard(NamedTuple):
    """Card definition class."""

    head: str
    body: str
    footer: Optional[str] = None
    classes: Optional[str] = None


class ICardClasses(IClasses):
    """Cards html classes."""

    container: str = cls("container", "px-0")
    card: str = cls("card", "px-0", "mb-4")
    card_body: str = cls("card-body")
    card_heading: str = cls("card-header")
    card_text: str = cls("card-text")
    card_footer: str = cls("card-footer small")


class Cards(Snippet):
    """A styled section tag."""

    TAG = 'div'

    def __init__(
        self,
        columns: Optional[int] = 2,
        items: List[ICard] = [],
        classes: ICardClasses = ICardClasses(),
    ) -> None:
        """Create the grid layout."""
        super().__init__(
            styles=None,
            classes=classes,
            className=classes.container)

        with self:
            self.items = Grid(columns)

        with self.items:
            for item in items:
                self.add_card(item.head, item.body, item.footer, item.classes)

    def add_card(
        self,
        head: str,
        body: str,
        footer: Optional[str] = None,
        card_cls: Optional[str] = None,
        head_tag: Type[html_tag] = h5,
        body_tag: Type[html_tag] = p
    ) -> None:
        """Add cards to the grid."""
        with self.items:
            with div(
                    className=cls(
                        self.classes.card,
                        card_cls if card_cls is not None else '')):
                head_tag(head, className=self.classes.card_heading)
                with div(className=self.classes.card_body):
                    body_tag(body, className=self.classes.card_text)
                if footer is not None:
                    div(footer, className=self.classes.card_footer)



"""Base class for eCharts Options Model."""

import numpy as np
import pandas as pd
from pydantic import BaseModel as PyDanticBase, Extra

from ezcharts.plots.util import JSCode

json_encoders = {
    pd.core.indexes.base.Index: lambda x: x.to_list(),
    pd.Series: lambda x: x.to_json(),
    np.ndarray: lambda x: x.tolist(),
    np.number: lambda x: float(x),
    JSCode: lambda x: x.to_json()}


class BaseModel(
        PyDanticBase, extra=Extra.forbid, validate_assignment=True,
        json_encoders=json_encoders, arbitrary_types_allowed=True):
    """Base class for eCharts Options."""
    # Note: this class's name in referenced in generate-model.py
    # TODO: can we overide getting of attributes to allow
    #       allow setting of children when parent isn't set.
    #       Simple __getattr__ isn't called :(



"""Relational plots."""

import numbers

from seaborn.relational import _LinePlotter, _ScatterPlotter

from ezcharts.plots import Plot, util


__all__ = ["relplot", "scatterplot", "lineplot"]


def relplot(
        data=None, *,
        x=None, y=None, hue=None, size=None, style=None, units=None,
        row=None, col=None, col_wrap=None, row_order=None, col_order=None,
        palette=None, hue_order=None, hue_norm=None,
        sizes=None, size_order=None, size_norm=None,
        markers=None, dashes=None, style_order=None,
        legend="auto", kind="scatter", height=5, aspect=1, facet_kws=None,
        **kwargs):
    """Figure-level interface for drawing relational plots onto a FacetGrid."""
    raise NotImplementedError


class Mixin:
    """Mixin class to define how to plot lines and points."""

    def plot(self, plt: Plot, kws):
        """Plot the plot."""
        data = self.plot_data.dropna()
        if data.empty:
            return

        if 'hue' not in data:
            data['hue'] = None

        x_name = self.variables.get("x", None)
        y_name = self.variables.get("y", None)

        plt.xAxis = dict(
            name=x_name,
            type=util.sns_type_to_echarts[str(self.var_types['x'])])
        plt.yAxis = dict(
            name=y_name,
            type=util.sns_type_to_echarts[str(self.var_types['y'])])

        plt.add_dataset({
            'id': 'raw',
            'dimensions': data.columns,
            'source': data.values})
        # TODO: size and style are also grouping "semantic" variables

        for series_index, series_name in enumerate(
                data['hue'].unique(), start=1):
            plt.add_dataset({
                'id': str(series_name),
                'fromDatasetId': 'raw',
                'transform': [{
                    'type': 'filter',
                    'config': {'dimension': 'hue', '=': series_name}}]})
            plt.add_series({
                'type': self.series_type,
                'name': series_name,
                # would be nicer to use datasetId here, but not documented
                # so not in our API
                'datasetIndex': series_index,
                'encode': {
                    'x': x_name, 'y': y_name,
                    'itemName': x_name, 'tooltip': [y_name]}})
            plt.tooltip = {"trigger": "axis"}

            # They keyword parameters are optionally set here so as not to override
            # theme defaults.
            for series in plt.series:
                if kws.get('s'):
                    if isinstance(kws['s'], numbers.Number):
                        series.symbolSize = kws['s']
                    else:
                        raise NotImplementedError(
                            "symbol size (s) currently only accepts a constant value")

                # marker/symbol options are:
                # circle, rect, roundRect, triangle, diamond, pin, arrow, none
                if kws.get('marker'):
                    series.symbol = kws['marker']

        # TODO: add legend
        return plt


class ScatterPlotter(Mixin, _ScatterPlotter):
    """Making scatter plots."""

    series_type = 'scatter'


def scatterplot(
        data=None, *,
        x=None, y=None, hue=None, size=None, style=None,
        palette=None, hue_order=None, hue_norm=None,
        sizes=None, size_order=None, size_norm=None,
        markers=True, style_order=None, legend="auto", ax=None,
        **kwargs):
    """Draw a scatter plot with possibility of several semantic groupings."""
    # see https://github.com/mwaskom/seaborn/blob/949dec3666ab12a366d2fc05ef18d6e90625b5fa/seaborn/relational.py#L726  # noqa

    # TODO: what of this do we actually care for
    variables = ScatterPlotter.get_semantics(locals())
    p = ScatterPlotter(data=data, variables=variables, legend=legend)
    p.map_hue(palette=palette, order=hue_order, norm=hue_norm)
    p.map_size(sizes=sizes, order=size_order, norm=size_norm)
    p.map_style(markers=markers, order=style_order)

    plt = Plot()
    p.plot(plt, kwargs)

    return plt


class LinePlotter(Mixin, _LinePlotter):
    """Making line plots."""

    series_type = 'line'


def lineplot(
        data=None, *,
        x=None, y=None, hue=None, size=None, style=None, units=None,
        palette=None, hue_order=None, hue_norm=None,
        sizes=None, size_order=None, size_norm=None,
        dashes=True, markers=None, style_order=None,
        estimator="mean", errorbar=("ci", 95), n_boot=1000, seed=None,
        orient="x", sort=True, err_style="band", err_kws=None,
        legend="auto", ci="deprecated", ax=None, **kwargs):
    """Draw a line plot with possibility of several semantic groupings."""
    # see https://github.com/mwaskom/seaborn/blob/949dec3666ab12a366d2fc05ef18d6e90625b5fa/seaborn/relational.py#L597  # noqa

    # TODO: what of this do we actually care for
    variables = _LinePlotter.get_semantics(locals())
    p = LinePlotter(
        data=data, variables=variables, estimator=estimator, n_boot=n_boot,
        seed=seed, errorbar=errorbar, sort=sort, orient=orient,
        err_style=err_style, err_kws=err_kws, legend=legend)
    p.map_hue(palette=palette, order=hue_order, norm=hue_norm)
    p.map_size(sizes=sizes, order=size_order, norm=size_norm)
    p.map_style(markers=markers, dashes=dashes, order=style_order)

    plt = Plot()
    p.plot(plt, kwargs)
    return plt



"""Get default stats layouts."""
from typing import List, Optional, Tuple, Type

from dominate.tags import div, h3, html_tag, p

from ezcharts.layout.base import IClasses, Snippet
from ezcharts.layout.snippets import Grid
from ezcharts.layout.util import cls


class IStatsClasses(IClasses):
    """Stats html classes."""

    container: str = cls("container", "px-0", "mb-5")
    item: str = cls(
        "shadow-sm", "container", "p-4", "bg-white",
        "border", "rounded")
    item_heading: str = cls("h5", "mb-0", "pb-3")
    item_value: str = cls("fs-2", "mb-0")


class Stats(Snippet):
    """A styled section tag."""

    TAG = 'div'

    def __init__(
        self,
        columns: Optional[int] = 2,
        items: List[Tuple[str, str]] = [],
        classes: IStatsClasses = IStatsClasses(),
    ) -> None:
        """Create table."""
        super().__init__(
            styles=None,
            classes=classes,
            className=classes.container)

        with self:
            self.items = Grid(columns)

        with self.items:
            for item in items:
                self.add_stats_item(item[0], item[1])

    def add_stats_item(
        self,
        value: str,
        heading: str,
        heading_tag: Type[html_tag] = h3
    ) -> None:
        """Add a badge to the banner."""
        with self.items:
            with div(className=self.classes.item):
                heading_tag(
                    heading, className=self.classes.item_heading)
                p(value, className=self.classes.item_value)



"""A base snippet to inherit from."""
from abc import ABC
from dataclasses import dataclass
import re
from uuid import uuid4

from dominate.tags import html_tag


@dataclass
class IDataClassMixin:
    """An inheritable dataclass."""

    ...


class IClasses(ABC, IDataClassMixin):
    """Snippet html classes."""

    ...


class IStyles(ABC, IDataClassMixin):
    """Snippet html styles."""

    ...


class Snippet(html_tag):
    """Base snippet."""

    TAG: str = 'div'

    def __init__(
        self,
        *args,
        styles,
        classes,
        **kwargs
    ):
        """Create snippet."""
        self.styles = styles
        self.classes = classes
        self.uid = self.get_uid(self.__class__.__name__)

        kwargs.update({'tagname': self.TAG})
        if not kwargs.get('id'):
            kwargs['id'] = self.uid

        super().__init__(*args, **kwargs)

    def get_uid(self, name: str) -> str:
        """Generate an ID."""
        # make sure `name` does not contain special characters so that the resulting uid
        # can be safely used in HTML `id` attributes
        if not re.match(r"^[a-zA-Z0-9_\-]+$", name):
            raise ValueError(
                "`name` cannot contain characters other than"
                "alphanumeric, '_', and '-'."
            )
        return name + '_' + str(uuid4()).replace("-", "")



"""Get ezchart containers."""
import json

from dominate.tags import script
from dominate.util import raw

from ezcharts.layout.base import Snippet
from ezcharts.layout.util import render_template


class EZChart(Snippet):
    """Wraps an ezchart plot in a div."""

    TAG = 'div'

    def __init__(
        self,
        plot,
        theme: str,
        width: str = '100%',
        height: str = '500px'
    ) -> None:
        """Create a div and script tag for initialising the plot."""
        width = "100%"
        super().__init__(
            styles=None,
            classes=None,
            className="chart-container",
            style=f"width:{width}; height:{height};")

        with self:

            script(raw(render_template(
                """
                var chart_{{ id }} = echarts.init(
                    dom=document.getElementById('{{ id }}'),
                    theme='{{ t }}',
                    opts={renderer: '{{ c.renderer }}'});
                var opt_{{ id }} = {{ j | replace('"',"'") | safe }};
                chart_{{ id }}.setOption(opt_{{ id }});
                {% if w.endswith('%') %}
                    window.addEventListener('resize', function(){
                        chart_{{ id }}.resize();
                    })
                {% endif %}
                chart_{{ id }}.resize()
                """,
                c=plot, j=plot.to_json(), t=theme,
                w=width, id=self.uid)))


class EZChartTheme(script):
    """Loads an echarts theme in a script tag."""

    def __init__(
        self,
        theme: str
    ) -> None:
        """Load theme."""
        super().__init__(tagname='script')
        with self:
            raw(render_template(
                """
                echarts.registerTheme(
                    '{{ n.themeName }}',
                    {{ t | replace(\'"\',"\'") | safe }}.theme);
                """,
                n=theme, t=json.dumps(theme)))


# No module demo here as method is directly on plot class.



"""Get default tabs layouts."""
from typing import Optional
import warnings

from dominate.tags import button, div, li, style, ul
from dominate.util import raw

from ezcharts.layout.base import IClasses, IStyles, Snippet
from ezcharts.layout.util import cls, css, render_template


warnings.simplefilter('always', DeprecationWarning)


class ITabsClasses(IClasses):
    """Tabs html classes."""

    container: str = cls("justify-content-center")
    tab_buttons_list: str = cls("nav", "nav-tabs", "mb-2")
    tab_buttons_list_item: str = cls("nav-item")
    tab_button: str = cls("nav-link", "px-0", "me-4", "text-muted")
    tab_contents_container: str = cls("tab-content", "p-5")
    tab_content: str = cls("tab-pane", "fade", "show")


class ITabsStyles(IStyles):
    """Tabs inline css styles."""

    tab_button: str = css(
        "margin-bottom: -1px",
        "font-weight: 600",
        "cursor: pointer",
        "border-color: transparent")
    tab_button_active: str = css(
        "border-bottom: 2px solid #0079a4",
        "color: #0079a4!important")


class Tabs(Snippet):
    """A styled div tag for creating tabbed layouts."""

    TAG = 'div'
    BUTTON_TEMPLATE = "#{{ id }}>ul>li>button { {{ styles }} };"
    BUTTON_TEMPLATE_ACTIVE = "#{{ id }}>ul>li>button.active { {{ astyles }} };"

    def __init__(
        self,
        styles: ITabsStyles = ITabsStyles(),
        classes: ITabsClasses = ITabsClasses()
    ) -> None:
        """Create tabs component."""
        super().__init__(
            styles=styles,
            classes=classes,
            className=classes.container)
        # store the number of tabs so that we know if a tab is the first one and should
        # be active
        self._ntabs: int = 0
        with self:
            style(raw(render_template(
                self.BUTTON_TEMPLATE,
                id=self.uid,
                styles=self.styles.tab_button)))
            style(raw(render_template(
                self.BUTTON_TEMPLATE_ACTIVE,
                id=self.uid,
                astyles=self.styles.tab_button_active)))
            self.buttons = ul(
                className=classes.tab_buttons_list,
                id="pills-tab",
                role="tablist")
            self.contents = div(
                className=classes.tab_contents_container,
                id="pills-tabContent")

    @property
    def _active_class_name(self) -> str:
        return "" if self._ntabs > 0 else " active"

    def add_tab(self, title: str, active: Optional[bool] = None) -> div:
        """Add a tab button and content container."""
        if active is not None:
            warnings.warn(
                "Explicitly passing 'active' parameter is deprecated "
                "and will be ignored.",
                DeprecationWarning, stacklevel=2)

        tab_id = f"{self.uid}-{self.get_uid('Tab')}"
        with self.buttons:
            with li(
                className=self.classes.tab_buttons_list_item,
                role="presentation"
            ):
                button(
                    title,
                    className=f"{self.classes.tab_button}{self._active_class_name}",
                    id=f"{tab_id}-tab",
                    type="button",
                    role="tab",
                    **self._get_button_data_aria(tab_id))

        with self.contents:
            return self._build_tab_container(tab_id)

    def add_dropdown_menu(
        self, title: str = "Dropdown", change_header: bool = True
    ) -> ul:
        """Get a dropdown menu."""
        button_id = f"{self.uid}-{self.get_uid('DropdownMenu')}"
        with self.buttons:
            with li(
                className=self.classes.tab_buttons_list_item + ' dropdown',
                role="presentation"
            ):
                button(
                    title,
                    className=f"{self.classes.tab_button} dropdown-toggle",
                    id=button_id,
                    href="#",
                    **{"data-bs-toggle": "dropdown"},
                    **({"data-ezc-updateDropDownTitle": ""} if change_header else {}))
                return ul(className="dropdown-menu")

    def add_dropdown_tab(
            self, title: str, active: Optional[bool] = None) -> div:
        """Get a dropdown tab."""
        if active is not None:
            warnings.warn(
                "Explicitly passing 'active' parameter is deprecated "
                "and will be ignored.",
                DeprecationWarning, stacklevel=2)

        tab_id = f"{self.uid}-{self.get_uid('DropdownItem')}"
        with li():
            button(
                title,
                className=f"dropdown-item{self._active_class_name}",
                id=f"#{tab_id}-tab",
                type="button",
                role="tab",
                **self._get_button_data_aria(tab_id))

        with self.contents:
            return self._build_tab_container(tab_id)

    def _build_tab_container(self, tab_id: str) -> div:
        """Get tab content container."""
        tab_container = div(
            className=f"{self.classes.tab_content}{self._active_class_name}",
            id=tab_id,
            role="tabpanel",
            **{"aria-labelledby": f"{tab_id}-tab"})
        # increment the number of tabs and return the div
        self._ntabs += 1
        return tab_container

    def _get_button_data_aria(self, tab_id: str) -> dict:
        """Get ARIA data for tab buttons."""
        return {
            "data-bs-toggle": "tab",
            "data-bs-target": f"#{tab_id}",
            "aria-controls": f"{tab_id}",
        }



"""Useful reusable functions."""
import json
import os
from typing import Dict

from dominate.tags import script, style
from dominate.util import raw, text
from jinja2 import BaseLoader, Environment
from pkg_resources import resource_filename
import sass


def load_json(
    path: str
) -> Dict:
    """Return json file as a dict."""
    with open(path, 'r', encoding='utf-8') as content:
        return json.load(content)


def inline(
    path: str
) -> text:
    """Return a file as a string."""
    with open(path, 'r', encoding='utf-8') as content:
        return raw(content.read())


def resolve_import(path):
    """Find resource and return as list."""
    resolved = resource_filename('ezcharts', path)
    # Temporary measure
    if os.path.exists(resolved) and not os.path.isdir(resolved):
        return [[resolved]]
    return [[path]]


def transpile(path):
    """Compile scss to css."""
    compiled = sass.compile(
        filename=path,
        output_style='compressed',
        importers=[(0, resolve_import)])
    return raw(compiled)


def inline_script(path):
    """Inline a script from path."""
    return script(inline(path))


def transpile_style(path):
    """Transpile a stylesheet from path."""
    return style(transpile(path))


def write_report(path, document):
    """Write a report to file."""
    with open(path, 'w', encoding='utf-8') as out:
        out.write('<!DOCTYPE html>')
        out.write(document.render())


def render_template(template, **kwargs):
    """Render a jinja2 template."""
    rtemplate = Environment(
        loader=BaseLoader()).from_string(template)
    return rtemplate.render(**kwargs)


def cls(*classes: str) -> str:
    """Collect element classes from a list."""
    return ' '.join(classes)


def css(*styles: str) -> str:
    """Collect inline element styles from a list."""
    return ' '.join(s.strip(';') + ';' for s in styles)



"""Re-usable report components."""
from typing import List, Type

from dominate.tags import body, footer, head, header, main, title

from ezcharts.layout.base import Snippet
from ezcharts.layout.resource import (
    base_body_resources, base_head_resources, Resource)
from ezcharts.layout.snippets.document import DefaultBody, DefaultHead
from ezcharts.layout.util import write_report


class Report(Snippet):
    """A basic report."""

    TAG: str = 'html'

    def __init__(
        self,
        report_title,
        head_tag: Type[head] = DefaultHead,
        body_tag: Type[body] = DefaultBody,
        head_resources: List[Resource] = base_head_resources,
        body_resources: List[Resource] = base_body_resources
    ) -> None:
        """Create tag."""
        super().__init__(
            styles=None,
            classes=None)

        with self:
            # Adds the generic meta tags for us!
            self.head = head_tag()
            # Enables scroll spy globals for us!
            self.body = body_tag()

        with self.head:
            title(report_title)
            for resource in head_resources:
                resource()

        with self.body:
            self.header = header()
            self.main = main()
            self.footer = footer()
            for resource in body_resources:
                resource()

    def write(self, path):
        """Write a report to file."""
        write_report(path, self)



"""Categorical plots."""


from ezcharts.plots import Plot

__all__ = [
    "catplot", "stripplot", "swarmplot", "boxplot", "violinplot",
    "boxenplot", "pointplot", "barplot", "countplot"]


def catplot(*args, **kwargs):
    """Figure-level interface for drawing categorical plots in a grid."""
    raise NotImplementedError


def stripplot(*args, **kwargs):
    """Draw a scatterplot where one variable is categorical."""
    raise NotImplementedError


def swarmplot(*args, **kwargs):
    """Draw a categorical scatterplot with non-overlapping points."""
    raise NotImplementedError


def boxplot(*args, **kwargs):
    """Draw a box plot to show distributions with respect to categories."""
    raise NotImplementedError


def violinplot(*args, **kwargs):
    """Draw a combination of boxplot and kernel density estimate."""
    raise NotImplementedError


def boxenplot(*args, **kwargs):
    """Draw an enhanced box plot for larger datasets."""
    raise NotImplementedError


def pointplot(*args, **kwargs):
    """Show point estimates and confidence intervals using scatter plot."""
    # NOTE: default estimator is mean
    raise NotImplementedError


def barplot(
    data=None, *, x=None, y=None, hue=None, order=None, hue_order=None,
    estimator='mean', errorbar=('ci', 95), n_boot=1000, units=None, seed=None,
    orient=None, color=None, palette=None, saturation=0.75, width=0.8,
    errcolor='.26', errwidth=None, capsize=None, dodge=True, ci='deprecated',
        ax=None, **kwargs):
    """Show point estimates and confidence intervals as rectangular bars.

    Contrary to the seaborn implementation, setting `dodge=False` does not
    result in overlaying the bars, but rather stacking them.
    """
    x_name = x if x else data.columns[0]
    y_name = y if y else data.columns[1]

    plt = Plot()

    if hue not in data:
        data = data[[x_name, y_name]]
    else:
        # Transform to [x, hue, hue, hue ...]
        data = data.pivot(
            columns=hue, values=y_name, index=x_name).reset_index(drop=False)
        plt.legend = dict(orient='horizontal', top=25)

    if order:
        mapper = {x: pos for (pos, x) in enumerate(order)}
        data = data.sort_values(x_name, key=lambda x: x.map(mapper))

    plt.xAxis = dict(name=x_name, type='category')
    plt.yAxis = dict(name=y_name, type='value')

    plt.add_dataset({
        'id': 'raw',
        'dimensions': data.columns,
        'source': data.values})

    # Automatically map series to columns in the dataset
    if dodge:
        plt.series = [{'type': 'bar'}] * (data.shape[1] - 1)
    else:
        plt.series = [{'type': 'bar', 'stack': x_name}] * (data.shape[1] - 1)

    return plt


def countplot(*args, **kwargs):
    """Show the counts of observations in each categorical bin using bars."""
    raise NotImplementedError



"""A reusable workflow params table."""
import argparse
import json
import os

from pkg_resources import resource_filename

from ezcharts.components.reports.comp import ComponentReport
from ezcharts.layout.snippets.table import DataTable


class ParamsTable(DataTable):
    """Display workflow params."""

    def __init__(
        self,
        params_file: str,
        **kwargs
    ) -> None:
        """Create table."""
        if not os.path.isfile(params_file):
            raise IOError('`params` should be a JSON file.')
        super().__init__(['Key', 'Value'], **kwargs)

        with open(params_file, encoding='utf-8') as f:
            data = json.load(f)
            with self.body:
                for k, v in data.items():
                    self.add_row(title=None, columns=[k, str(v)])


def main(args):
    """Entry point to demonstrate a parameter table."""
    comp_title = 'Parameter Table'
    param_table = ParamsTable(args.params)
    report = ComponentReport(comp_title, param_table)
    report.write(args.output)


def argparser():
    """Argument parser for entrypoint."""
    parser = argparse.ArgumentParser(
        'Params table',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        add_help=False
    )
    parser.add_argument(
        "--params",
        default=resource_filename('ezcharts', "test/params.json"),
        help=(
            "A JSON file containing the workflow parameter "
            "key/values."
        )
    )
    parser.add_argument(
        "--output",
        default="params_table_report.html",
        help="Output HTML file."
    )
    return parser



"""Regression plots."""

__all__ = ["lmplot", "regplot", "residplot"]


def lmplot(*args, **kwargs):
    """Plot data and regression model fits across a FacetGrid."""
    raise NotImplementedError


def regplot(*args, **kwargs):
    """Plot data and a linear regression model fit."""
    raise NotImplementedError


def residplot(*args, **kwargs):
    """Plot the residuals of a linear regression."""
    raise NotImplementedError



"""Get default grid layout."""
from typing import Optional

from dominate.tags import style

from ezcharts.layout.base import IStyles, Snippet
from ezcharts.layout.util import css, render_template


class IGridStyles(IStyles):
    """Section inline css styles."""

    container: str = css(
        "display: grid",
        "grid-template-rows: repeat(1, 1fr)",
        "grid-template-columns: repeat(1, 1fr)",
        "grid-column-gap: 10px",
        "grid-row-gap: 20px")


class Grid(Snippet):
    """A styled div tag for creating css-grids."""

    TAG = 'div'

    def __init__(
        self,
        columns: Optional[int] = 2,
        styles: IGridStyles = IGridStyles()
    ) -> None:
        """Create css grid."""
        # We need to place the style tags before the
        # grid container otherwise contents won't
        # respect the rules properly. Therefore
        # we must generate an ID up front.
        uid = self.get_uid(self.__class__.__name__)

        # Adding extra style tags prior to the element
        # allows writing css in a conventional way,
        # rather than inline, permitting media queries.
        style(render_template(
            template=(
                """
                #{{grid_id}} {
                    {{ styles }}
                    {% if c %}
                    grid-template-columns: repeat({{ c }}, 1fr)
                    {% endif %}
                };
                """
            ),
            grid_id=uid,
            styles=styles.container,
            c=columns))

        # For some reason we have to break these up
        style(render_template(
            template=(
                """
                @media screen and (max-width: 800px) {
                    body #{{grid_id}} {
                        grid-template-columns: repeat(1, 1fr)
                    }
                }
                """
            ),
            grid_id=uid,
            styles=styles.container,
            c=columns))

        # Finally we call super to create the grid container
        # itself.
        super().__init__(id=uid, styles=styles, classes=None)
